#include "GLStuff.h"
#include "insaniac.h"

#include <cmath>

namespace GLStuff
{

   Slider gPitchMinSlider(20,20,100,20, kPitchMinParamId);
   Slider gPitchMaxSlider(140, 20, 100, 20, kPitchMaxParamId);
   Slider gLengthMinSlider(20,60,100,20, kSizeMinParamId);
   Slider gLengthMaxSlider(140, 60, 100, 20, kSizeMaxParamId);
   Slider gDensitySlider(280, 60, 100, 20, kDensityParamId);
   ToggleButton gToggle(280, 20, 20, 20);
   
   std::deque<GLDisplay*> gDisplays;
   
//function to draw the circle
void createcircle (int k, int r, int h) {
   
   glLineWidth(4);
   glBegin(GL_LINE_STRIP);
   
   int numSegments = 30;
   float inc = 2*__PI / (float)(numSegments-1);
   float pos = 0;
   for (int i = 0;i < numSegments;i++)
   {
		circle.x = r * cos(pos);
		circle.y = r * sin(pos);
		glVertex3f(circle.x,circle.y,0);
      
      pos += inc;
		
		//circle.x = r * cos(i) - h;
		//circle.y = r * sin(i) + k;
		//glVertex3f(circle.x + k,circle.y - h,0);
      
      
   }
   glEnd();
}

//-----------------------------------------------------------------------------
// name: rand2f()
// desc: generate a random float
//-----------------------------------------------------------------------------
GLfloat rand2f( float a, float b )
{
   return a + (b-a)*(rand() / (GLfloat)RAND_MAX);
}

//default constructor
Ball::Ball() {
	xpos = 0.0;
	ypos = 0.0;
	diameter = 0.0;
	id = 0;
	others = NULL;
	gravityON = false;
	bounces = 0;
   fCloudEvent = NULL;
   yposAtTimeOfCombination = 0.f;
}

//main constructor
Ball::Ball(float xin, float yin, float din, int idin, Ball* oin, bool gravityin, int bouncesin) {
	xpos = xin;
	ypos = yin;
	diameter = din;
	id = idin;
	others = oin;
	gravityON = gravityin;
	bounces = bouncesin;
   fCloudEvent = NULL;
}

void Ball::collide() {
   int j;
   for (int i = id + 1; i < id + numBalls; i++) {
      if (i >= numBalls) j = i-numBalls;
      else j = i;
      if (gravityON == true) {
         float dxpos = others[j].xpos - xpos;
         float dypos = others[j].ypos - ypos;
         float distance = sqrt(dxpos*dxpos + dypos*dypos);
         float minDist = others[j].diameter/2 + diameter/2;
         if (distance < minDist && others[j].diameter > 0) {
            float angle = atan2(dypos, dxpos);
            float targetX = xpos + cos(angle) * minDist;
            float targetY = ypos + sin(angle) * minDist;
            float axpos = (targetX - others[j].xpos) * spring;
            float aypos = (targetY - others[j].ypos) * spring;
            vxpos -= axpos;
            vypos -= aypos;
            others[j].vxpos += axpos;
            others[j].vypos += aypos;
         }
      }
   }   
}

void Ball::combine() {
   int j;
   for (int i = id + 1; i < id + numBalls; i++) {
      if (i >= numBalls) j = i-numBalls;
      else j = i;
      float dxpos = others[j].xpos - xpos;
      float dypos = others[j].ypos - ypos;
      float distance = sqrt(dxpos*dxpos + dypos*dypos);
      float minDist = others[j].diameter/2 + diameter/2;
      if (diameter+others[j].diameter <= g_diameter*g_combineBeforeDrop) {
         if (distance < minDist) { 
            if (diameter >= others[j].diameter ) {
            
               if (others[j].diameter != 0)
               {
                  Insaniac::GetInstance()->AddNoiseBlip();
                  std::cout << "create blip\n";
               }
               diameter += others[j].diameter;
               others[j].diameter = 0;
            }
            if (diameter < others[j].diameter) {
               
               Insaniac::GetInstance()->AddNoiseBlip();
               std::cout << "create blip\n";
            
               others[j].diameter += diameter;
               diameter = 0;
            }

         }
      }
      
      if (diameter == g_diameter*g_combineBeforeDrop)
      {
         gravityON = true;
         if (!fCloudEvent) {
            fCloudEvent = new CloudEvent;
            Insaniac::GetInstance()->AddCloudEvent(fCloudEvent);
            yposAtTimeOfCombination = ypos;
         }
      }
      
      //if (others[j].diameter == g_diameter*g_combineBeforeDrop) others[j].gravityON = true;
      
      if (distance < minDist*(gWidth/g_diameter) && others[j].diameter != 0 && others[j].gravityON == false && gravityON == false) { 
         float angle = atan2(dypos, dxpos);
         float targetX = xpos + cos(angle) * minDist;
         float targetY = ypos + sin(angle) * minDist;
         float axpos = (targetX - others[j].xpos) * spring;
         float aypos = (targetY - others[j].ypos) * spring;
         float magnetScale = diameter*g_speed/(1000*pow(distance,2));
         vxpos -= axpos*magnetScale/diameter;
         vypos -= aypos*magnetScale/diameter;
         others[j].vxpos += axpos*magnetScale;
         others[j].vypos += aypos*magnetScale;
      }              
   }   
}


int gReporter = 0;

void Ball::move() {
   
   if (gravityON) vypos -= gravity;   
   vxpos *= friction;
   vypos *= friction;
   xpos += vxpos;
   ypos += vypos;
   if (xpos + diameter/2 > gWidth) {
      xpos = gWidth - diameter/2;
      vxpos *= -1; 
   }
   else if (xpos - diameter/2 < -gWidth) {
      xpos = -gWidth + diameter/2;
      vxpos *= -1;
   }
   if (ypos + diameter/2 > gHeight) {
      if (bounces < g_bounces) {
         ypos = gHeight - diameter/2;
         vypos *= -1;
         bounces++;
      }
      else {
         vypos = 0;
         ypos = rand()%(gHeight/4)+1;
         xpos = rand()%(gWidth)+1;
         diameter = g_diameter;
         gravityON = false;

         if (fCloudEvent)
            Insaniac::GetInstance()->RemoveCloudEvent(fCloudEvent);
         bounces = 0; 
      }
   }
   else if (ypos - diameter/2 < -gHeight) {
      ypos = -gHeight + diameter/2;
      vypos *= -1;
   }
  
   if (gravityON && !gReporter)
      gReporter = id;
      
   if (gravityON && this->id == gReporter)
      ;//std::cout << ypos << std::endl;
      
   if (gravityON && fCloudEvent)
   {
      float gain = 1 - (600 - ypos) / (600 - yposAtTimeOfCombination);
      if (gain > 0.5)
         gain = 1-gain;
         
      gain *= 0.4;

      fCloudEvent->SetGain(gain);
      if (this->id == gReporter)
         ;//std::cout << gain << std::endl;
   }

}

Ball::~Ball()
{
   if (fCloudEvent)
   {
      Insaniac::GetInstance()->RemoveCloudEvent(fCloudEvent);
      delete fCloudEvent;
      fCloudEvent = NULL;
   }
}

void Ball::display() {
   if (diameter > 0) {
		glColor3f(1.0/id,0.0,std::max(1.1/id, 0.6));
      glPushMatrix();
		glTranslatef(xpos,-ypos,0);
		//glutSolidSphere(diameter/2, 10, 10);
		createcircle(0,diameter/2,0);
      glPopMatrix();
   }
}

void UpdateControls()
{
   //gStaggered = gToggle.Toggled();
   //gLogFactor = (1 - gPitchMinSlider.Value()) * 0.75 + 0.25;
   //gSamplesPerFrame = gPitchMaxSlider.Value() * 40000 + 512;
}


void GLStartup(int argc, const char** argv)
{
	std::cout << "Hit q to quit.\n";
 
   numBalls = 150; //sets the number of balls
   g_speed = 12; //sets the speed of the balls
   g_bounces = 8; //the number of bounces before the gravitied balls disappear
   gravity = -.4; //the degree of gravity (must be negative)
   friction = .98, spring = 0.2; //just leave these as they are
   g_diameter = 14.3; //ball diameter
   g_combineBeforeDrop = 4; //number of times combinations take place before the balls drop
   g_X = 0.0; g_Y = 0.0; g_ZOOM = 1500; //initial states for the gl_view
   // width and height of the window
   gWidth = 800;
   gHeight = 600;
   gLastWidth = gWidth;
   gLastHeight = gHeight;
   // light 0 position
   g_light0_pos[0] = 2.0f;
   g_light0_pos[1] =  8.2f;
   g_light0_pos[2] = 4.0f;
   g_light0_pos[3] = 1.0f;
   
   balls = new Ball[numBalls];
   gDisplay = true;
   gFullscreen = false;
   gBufferSize = 1024;
   gFrameRate = 40; // frames/sec
   gRefreshRate = 1000 / gFrameRate; // ms
   gMenuItem = 0;
   gInc_val_mouse = 360 / gFrameRate;
   gLeftButtondown = false;
   gMouseOriginX = 0;
   gMouseOriginY = 0;
   // experimental
   gHopCount = 0;
   gRings = false;
   gHop = 10;
   gShowHUD = true;
   
//   Slider gPitchMinSlider(20,20,100,20, kPitchMinParamId);
//   Slider gPitchMaxSlider(140, 20, 100, 20, kPitchMaxParamId);
//   Slider gLengthMinSlider(20,60,100,20, kSizeMinParamId);
//   Slider gLengthMaxSlider(140, 60, 100, 20, kSizeMaxParamId);
//   Slider gDensitySlider(280, 60, 100, 20, kDensityParamId);
//   ToggleButton gToggle(280, 20, 20, 20);
//   
//   std::deque<GLDisplay*> gDisplays;
   
   gPitchMinSlider.SetValue(0.2);
   gPitchMaxSlider.SetValue(0.5);
   gLengthMinSlider.SetValue(0.2);
   gLengthMaxSlider.SetValue(0.5);
   gDensitySlider.SetValue(0.5);
   

   
   
#ifdef __OS_MACOSX__
   //save working dir
   char * cwd = getcwd( NULL, 0 );
#endif
   // initialize GLUT
   glutInit( &argc, (char**)argv );
   
#ifdef __OS_MACOSX__
   //restore working dir
   chdir( cwd );
   free( cwd );
#endif
   
   glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
   glutInitWindowSize( gWidth, gHeight );
   glutInitWindowPosition( 100, 100 );
   glutCreateWindow( "Zomg!" );
   //glutIdleFunc( idleFunc );
   glutDisplayFunc( displayFunc );
   glutReshapeFunc( reshapeFunc );
   glutKeyboardFunc( keyboardFunc );
   glutMouseFunc( mouseFunc );
   glutMotionFunc(motionFunc);        
   glutTimerFunc(gRefreshRate,timerCallback,0);
   createMenu();
   if( gFullscreen )
      glutFullScreen();
   
   // do our own initialization
   initialize_graphics( );
   
   glutMainLoop();
   
}

void initialize_graphics()
{
   // set the GL clear color - use when the color buffer is cleared
   glClearColor( 0, 0,0, 1.0f );
   // set the shading model to 'smooth'
   glShadeModel( GL_SMOOTH );
   // enable depth
   //glEnable( GL_DEPTH_TEST );
   // set the front faces of polygons
   glFrontFace( GL_CCW );
   // draw in wireframe
   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	
	glEnable( GL_LINE_SMOOTH );
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	
	glEnable(GL_BLEND);              //activate blending mode
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLineWidth(4.0);
	
	
	// enable lighting
   glEnable( GL_LIGHTING );
   // enable lighting for front
   glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
   // material have diffuse and ambient lighting 
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
   // enable color
   glEnable( GL_COLOR_MATERIAL );
	
	// enable light 0
   glEnable( GL_LIGHT0 );    
   // set the position of the lights
   glLightfv( GL_LIGHT0, GL_POSITION, g_light0_pos );
   
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
   // save the new window size
   gWidth = w; gHeight = h;
   // map the view port to the client area
   glViewport( 0, 0, w, h );
   // set the matrix mode to project
   glMatrixMode( GL_PROJECTION );
   // load the identity matrix
   glLoadIdentity( );
   // create the viewing frustum
   gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 1550.0 );
   // set the matrix mode to modelview
   glMatrixMode( GL_MODELVIEW );
   // load the identity matrix
   glLoadIdentity( );
   // position the view point
   gluLookAt( g_X, g_Y, g_ZOOM, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}

void go2d() {
   
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluOrtho2D(0.0, gWidth, 0.0, gHeight);
   
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   
   glDisable(GL_TEXTURE);
   glDisable(GL_TEXTURE_2D);   
   
}

void go3d() {
   glEnable(GL_TEXTURE);
   glEnable(GL_TEXTURE_2D);
   
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
   
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   
   reshapeFunc(gWidth, gHeight);
}

//-----------------------------------------------------------------------------
// Name: generateBalls( )
// Desc: A function to create all of the balls in the visualization space
//-----------------------------------------------------------------------------
void generateBalls()
{
	for (int i = 0; i < numBalls; i++) {
		balls[i] = Ball(rand2f(-gWidth,gWidth),(rand2f(-gHeight,gHeight))/2, g_diameter, i, balls, false, 0);
		//balls[i] = Ball(grain_position[i].xpos, grain_position[i].ypos, g_diameter, i, balls, false, 0);
	}
}

void addBall()
{
   for (int i = 0; i < numBalls; i++) {
      if (balls[i].diameter == 0) {
         balls[i] = Ball(rand2f(-gWidth,gWidth),(rand2f(-gHeight,gHeight))/2, g_diameter, i, balls, false, 0);
         break;
      }
   }
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
   switch( key )
   {
      case 'f':
		{
			if( !gFullscreen )
			{
				gLastWidth = gWidth;
				gLastHeight = gHeight;
				glutFullScreen();
			}
			else
				glutReshapeWindow( gLastWidth, gLastHeight );
			
			gFullscreen = !gFullscreen;
		}
         break;
      case 'h':
         gShowHUD = !gShowHUD;
         break;
      case 'w':
         gFrameRate++;
         gRefreshRate = 1000 / gFrameRate;
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 's':
         if (gFrameRate > 1)
         {
            gFrameRate--;
            gRefreshRate = 1000 / gFrameRate;
         }
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 'e':
         break;
         
      case 'd':
         break;
      case '1':
      {
      }
         break;
      case '2':
      {
      }
         break;
		case '3':
      {
      }
			break;
      case '4':
      {
      }
			break;
      case 'r':
         break;
      case 'c':
         break;
         
		case 'q':
		{
			exit(0);
		}
			break;
		default:
			break;
   }
	
   // do a reshape since gEyeY might have changed
   reshapeFunc( gWidth, gHeight );
   glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
   if( button == GLUT_LEFT_BUTTON )
   {
      // rotate
      if( state == GLUT_DOWN )
      {
         gLeftButtondown = true;
         gMouseOriginX = x;
         gMouseOriginY = y;   
         
         if (gPitchMinSlider.Hit(x, y))
         {
            gPitchMinSlider.Touch(x, y);
         }
         
         if (gPitchMaxSlider.Hit(x, y))
         {
            gPitchMaxSlider.Touch(x, y);
         }
         
         if (gLengthMinSlider.Hit(x, y))
         {
            gLengthMinSlider.Touch(x, y);
         }
         
         if (gLengthMaxSlider.Hit(x, y))
         {
            gLengthMaxSlider.Touch(x, y);
         }
         
         if (gDensitySlider.Hit(x, y))
         {
            gDensitySlider.Touch(x, y);
         }
         
         if (gToggle.Hit(x,y))
         {
            gToggle.Toggle();
         }
      }
      else if (state == GLUT_UP)
      {
         gPitchMinSlider.UnTouch();
         gPitchMaxSlider.UnTouch();
         gLengthMinSlider.UnTouch();
         gLengthMaxSlider.UnTouch();
         gDensitySlider.UnTouch();
         gLeftButtondown = false;
      }
      // gInc -= gInc_val_mouse;
      else {
         //gInc += gInc_val_mouse;
         
         
      }
   }
   
   
   glutPostRedisplay( );
}

void motionFunc(int x, int y )
{   
   if (gLeftButtondown)
   {
      if (gPitchMinSlider.Touched())
      {
         gPitchMinSlider.Touch(x,y);
      }
      else if (gPitchMaxSlider.Touched())
      {
         gPitchMaxSlider.Touch(x,y);
      }
      else if (gLengthMinSlider.Touched())
      {
         gLengthMinSlider.Touch(x,y);
      }
      else if (gLengthMaxSlider.Touched())
      {
         gLengthMaxSlider.Touch(x,y);
      }
      else if (gDensitySlider.Touched())
      {
         gDensitySlider.Touch(x,y);
      }
      
      else
      {
         
      }
   }
   
   gMouseOriginX = x;
   gMouseOriginY = y;
}

void displayFunc( )
{
   
   static GLfloat x = 0.0f;
	
   // clear the color and depth buffers
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glPushMatrix();
	
	for (int i = 0; i < numBalls; i++) {
		if (balls[i].diameter != 0) {
			balls[i].combine();
			balls[i].collide();
			balls[i].move();
			balls[i].display();
		}
		//else balls[i] = Ball(rand2f(-gWidth,gWidth),(rand2f(-gHeight,gHeight))/2, g_diameter, i, balls, false, 0);
   }
	
	
	glPopMatrix();
   
   //UpdateControls();
   
   if(gShowHUD)
   {
      gPitchMinSlider.Draw();
      gPitchMaxSlider.Draw();
      gLengthMinSlider.Draw();
      gLengthMaxSlider.Draw();
      gDensitySlider.Draw();
      gToggle.Draw();
      
      glColor4f(1,1,1,1);
      std::string s = "Freq-";
      DrawString(gPitchMinSlider.X() - 20,gPitchMinSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      glColor4f(1,1,1,1);
      s = "Freq+";
      DrawString(gPitchMaxSlider.X() - 20,gPitchMaxSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      glColor4f(1,1,1,1);
      s = "Length-";
      DrawString(gLengthMinSlider.X() - 20,gLengthMinSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      glColor4f(1,1,1,1);
      s = "Length+";
      DrawString(gLengthMaxSlider.X() - 20,gLengthMaxSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      s = "Density";
      DrawString(gDensitySlider.X() - 20,gDensitySlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      
      glColor4f(1,1,1,1);
      s = "???";
      DrawString(gToggle.X() - 10,gToggle.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
      
      glLineWidth(1);
      s = "Insaniac 0.0.0.0.42.69";
      DrawString(gWidth - 500, gHeight-10, s, GLUT_BITMAP_TIMES_ROMAN_10, 0.07);
   }
   
   glFlush( );
   glutSwapBuffers( );
}

void timerCallback(int v)
{
   //gAngleY += gInc;
   glutPostRedisplay();
   glutTimerFunc(gRefreshRate,timerCallback,0);
}

void createMenu(void){
   //////////
   // MENU //
   //////////
   
   // Create a submenu, this has to be done first.
   int submenid = glutCreateMenu(menu);
   
   // Add sub menu entry
   glutAddMenuEntry("Teapot", 2);
   glutAddMenuEntry("Cube", 3);
   glutAddMenuEntry("Torus", 4);
   
   // Create the menu, this menu becomes the current menu
   int menid = glutCreateMenu(menu);
   
   // Create an entry
   glutAddMenuEntry("Clear", 1);
   
   glutAddSubMenu("Draw", submenid);
   // Create an entry
   glutAddMenuEntry("Quit", 0);
   
   // Let the menu respond on the right mouse button
   glutAttachMenu(GLUT_RIGHT_BUTTON);
   
   
}


void menu(int value){
   if(value == 0){
      exit(0);
   }else{
      gMenuItem = value;
   }
   
   // you would want to redraw now
   glutPostRedisplay();
}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
   // render the scene
   //glutPostRedisplay( );
}

void DrawString(int x, int y, std::string& s, void* font, float scaleFactor/* = 0.2f*/)
{
   go2d();
   
   glRasterPos3f(10, 10, 0);
   
   glTranslatef(x,y,0); 
   glScalef(scaleFactor,scaleFactor,1);
   
   glColor4f(1,1,1,1);
   for (std::string::iterator i = s.begin(); i != s.end(); ++i)
   {
      glutStrokeCharacter( GLUT_STROKE_ROMAN, *i );
      //glutBitmapCharacter(font, *i);
   }
   
   go3d();
}

}

