#ifndef h_ParameterAPI
#define h_ParameterAPI

#include <string>
#include <vector>

class Broadcaster;
class Observable;
class Parameter;

class Listener
{
public:
	virtual void Listen(Broadcaster const* b) = 0;
};

class ParameterListener : public Listener
{
public:
	virtual void Listen(Broadcaster const* b)
	{
		Parameter const* p = reinterpret_cast<Parameter const*> (b);
		this->Listen(p);
	}
	
	virtual void Listen(Parameter const* p) = 0;
};

class Broadcaster // : public PseudoSingleton (TODO)
{
public:
	void AddListener(Listener* l);
	void RemoveListener(Listener* l);
protected:
	virtual void Broadcast() = 0;
	std::vector<Listener*> _listeners;
};

class Observer
{
public:
	virtual void Observe(Observable* o) = 0;
};

class ParameterObserver : public Observer
{
public:
	virtual void Observe(Observable* o)
	{
		Parameter* p = reinterpret_cast<Parameter*> (o);
		this->Observe(p);	
	}
	
	virtual void Observe(Parameter* p) = 0;
	
};

class Observable
{
public:
	void AddObserver(Observer* o);
	void RemoveObserver(Observer* o);
protected:
	virtual void NotifyObservers() = 0;
	std::vector<Observer*> _observers;
};

class Parameter : public Broadcaster
				, public Observable
{
public:
	Parameter(const int id, const std::string& name, const float initialValue = 0.f);
	
	const float Value() const;
	const std::string Name() const;
	const std::string Text() const;
   const int Id() const;
	
	void SetValue(const float value);

protected:
	// override Observable
	virtual void NotifyObservers();
	
	// override Broadcast
	virtual void Broadcast();
private:
	int _id;
	float _value;
	std::string _name;
	
	bool _isPublished;
	bool _isBroadcast;
};

class ParameterManager // : public PsuedoSingleton (TODO)
{
public:

   ~ParameterManager();

   static ParameterManager* GetInstance()
	{
		if (!sInstance)
		{
			sInstance = new ParameterManager;
		}
		return sInstance;
	}

	void AddParameter(Parameter* p);
	Parameter* GetParameter(int id);
private:
   static ParameterManager* sInstance;

	std::vector<Parameter*> _parameters;
};

#endif
